@@ -276,7 +276,7 @@ where
             return Ok(());
         }
 
-        // Null visit checks all nested stats are reacheable without loops.
+        // Null visit checks all nested stats are reachable without loops.
         self.visit_meta(self.top.as_ref().unwrap(), &mut |_| Ok(()))
     }
 
@@ -678,7 +678,7 @@ where
     fn drop(&mut self) {
         self.exit.store(true, Ordering::Relaxed);
         if let Some(path) = self.path.as_ref() {
-            let _ = StatsClient::new().set_path(path).connect();
+            let _ = StatsClient::new().set_path(path).connect(None);
         }
     }
 }
