@@ -8,11 +8,15 @@ fn main() {
         .unwrap()
         .enable_intf("src/bpf/intf.h", "bpf_intf.rs")
         .enable_skel("src/bpf/main.bpf.c", "bpf")
+        .add_source("src/bpf/balance.bpf.c")
+        .add_source("src/bpf/idle.bpf.c")
         .add_source("src/bpf/introspec.bpf.c")
+        .add_source("src/bpf/lat_cri.bpf.c")
         .add_source("src/bpf/lock.bpf.c")
         .add_source("src/bpf/power.bpf.c")
         .add_source("src/bpf/preempt.bpf.c")
         .add_source("src/bpf/sys_stat.bpf.c")
+        .add_source("src/bpf/util.bpf.c")
         .add_source("src/bpf/lib/arena.bpf.c")
         .add_source("src/bpf/lib/atq.bpf.c")
         .add_source("src/bpf/lib/bitmap.bpf.c")
