@@ -4,56 +4,18 @@
  * Author: Changwoo Min <changwoo@igalia.com>
  */
 
-/*
- * To be included to the main.bpf.c
- */
-struct pick_ctx {
-	/*
-	 * Input arguments for pick_idle_cpu().
-	 */
-	const struct task_struct *p;
-	task_ctx *taskc;
-	u64 wake_flags;
-	s32 prev_cpu;
-	/*
-	 * Additional input arguments for find_sticky_cpu_and_cpdom().
-	 */
-	s32 sync_waker_cpu;
-	/*
-	 * Additional output arguments for init_active_ovrflw_masks().
-	 */
-	struct bpf_cpumask *active; /* global active mask */
-	struct bpf_cpumask *ovrflw; /* global overflow mask */
-	/*
-	 * Additional output arguments for init_ao_masks().
-	 * Additional input arguments for find_sticky_cpu_and_cpdom().
-	 */
-	struct cpu_ctx *cpuc_cur;
-	struct bpf_cpumask *a_mask; /* task's active mask */
-	struct bpf_cpumask *o_mask; /* task's overflow mask */
-	/*
-	 * Additional input arguments for init_idle_i_mask().
-	 */
-	const struct cpumask *i_mask;
-	/*
-	 * Additional input arguments for init_idle_ato_masks().
-	 * Additional input arguments for pick_idle_cpu_at_cpdom().
-	 */
-	struct bpf_cpumask *ia_mask;
-	struct bpf_cpumask *iat_mask;
-	struct bpf_cpumask *io_mask;
-	struct bpf_cpumask *temp_mask;
-	/*
-	 * Flags.
-	 */
-	bool a_empty:1;
-	bool o_empty:1;
-	bool is_task_big:1;
-	bool i_empty:1;
-	bool ia_empty:1;
-	bool iat_empty:1;
-	bool io_empty:1;
-};
+#include <scx/common.bpf.h>
+#include <bpf_arena_common.bpf.h>
+#include "intf.h"
+#include "lavd.bpf.h"
+#include "util.bpf.h"
+#include "power.bpf.h"
+#include <errno.h>
+#include <stdbool.h>
+#include <bpf/bpf_core_read.h>
+#include <bpf/bpf_helpers.h>
+#include <bpf/bpf_tracing.h>
+#include <lib/cgroup.h>
 
 struct sticky_ctx {
 	/*
@@ -127,7 +89,7 @@ bool init_idle_ato_masks(struct pick_ctx *ctx, const struct cpumask *idle_mask)
 	ctx->ia_mask = ctx->cpuc_cur->tmp_t_mask;
 	ctx->io_mask = ctx->cpuc_cur->tmp_t2_mask;
 	ctx->iat_mask = ctx->cpuc_cur->tmp_t3_mask;
-	ctx->temp_mask = ctx->cpuc_cur->tmp_l_mask; /* l_mask is no longer used, recyle it. */
+	ctx->temp_mask = ctx->cpuc_cur->tmp_l_mask; /* l_mask is no longer used, recycle it. */
 	if (!ctx->ia_mask || !ctx->io_mask || !ctx->iat_mask || !ctx->temp_mask)
 		return false;
 
@@ -155,7 +117,8 @@ bool init_idle_ato_masks(struct pick_ctx *ctx, const struct cpumask *idle_mask)
 	return true;
 }
 
-static s32 find_cpu_in(const struct cpumask *src_mask, struct cpu_ctx *cpuc_cur)
+__hidden
+s32 find_cpu_in(const struct cpumask *src_mask, struct cpu_ctx *cpuc_cur)
 {
 	const volatile u16 *cpu_order = get_cpu_order();
 	const struct cpumask *online_mask;
@@ -188,8 +151,8 @@ static s32 find_cpu_in(const struct cpumask *src_mask, struct cpu_ctx *cpuc_cur)
 	return -ENOENT;
 }
 
-static
-s32 pick_idle_cpu_at_cpdom(struct pick_ctx *ctx, s64 cpdom, u64 scope,
+
+static s32 pick_idle_cpu_at_cpdom(struct pick_ctx *ctx, s64 cpdom, u64 scope,
 			   bool *is_idle)
 {
 	struct bpf_cpumask *cpd_mask;
@@ -395,7 +358,7 @@ bool is_sync_wakeup(struct pick_ctx *ctx)
 	return true;
 }
 
-static 
+static
 s32 find_sticky_cpu_and_cpdom(struct pick_ctx *ctx, s64 *sticky_cpdom)
 {
 	struct cpu_ctx *p0, *p1, *cpuc;
@@ -533,7 +496,7 @@ bool is_sync_waker_idle(struct pick_ctx * ctx, s64 *cpdom_id)
 		return false;
 
 	cpuc_waker = get_cpu_ctx_id(ctx->sync_waker_cpu);
-	if (!cpuc_waker || nr_queued_on_cpu(cpuc_waker))
+	if (!cpuc_waker || queued_on_cpu(cpuc_waker))
 		return false;
 
 	if (nr_cpdoms > 1) {
@@ -548,31 +511,86 @@ bool is_sync_waker_idle(struct pick_ctx * ctx, s64 *cpdom_id)
 }
 
 static
-s32 __attribute__ ((noinline)) pick_idle_cpu(struct pick_ctx *ctx, bool *is_idle)
+s32 migrate_to_neighbor(struct pick_ctx *ctx, struct cpdom_ctx *cpdc,
+			u64 scope, s64 *sticky_cpdom, bool *is_idle)
+{
+	struct cpdom_ctx *mig_cpdc;
+	s64 mig_cpdom, nr_nbr;
+	s32 cpu = -ENOENT;
+	int i, j;
+
+	/*
+	 * Let's migrate a task to neighbor domain when:
+	 *  1) The sticky domain is over-loaded (cpdc->is_stealee)
+	 *  2) The target domain is under-loaded (mig_cpdc->is_stealer)
+	 *     that has a fully idle core.
+	 *
+	 * Note that when a system is under-loaded, task donation works better
+	 * than task stealing because DSQs are mostly empty (i.e., it is hard
+	 * to steal from a DSQ).
+	 */
+	bpf_for(i, 0, LAVD_CPDOM_MAX_DIST) {
+		nr_nbr = min(cpdc->nr_neighbors[i], LAVD_CPDOM_MAX_NR);
+		if (nr_nbr == 0)
+			break;
+		bpf_for(j, 0, LAVD_CPDOM_MAX_NR) {
+			if (j >= nr_nbr)
+				break;
+			mig_cpdom = get_neighbor_id(cpdc, i, j);
+			if (mig_cpdom < 0)
+				continue;
+
+			mig_cpdc = MEMBER_VPTR(cpdom_ctxs, [mig_cpdom]);
+			if (!mig_cpdc || !READ_ONCE(mig_cpdc->is_stealer))
+				continue;
+
+			cpu = pick_idle_cpu_at_cpdom(ctx, mig_cpdom, scope, is_idle);
+			if (cpu >= 0) {
+				/*
+				 * If task donation is successful, mark the stealer
+				 * and the stealee's job done. By marking done,
+				 * those compute domains would not be involved in
+				 * load balancing until the end of this round,
+				 * so this helps gradual migration. It is racy
+				 * in task stealings and donations, but we don't
+				 * care because a slight over-migration does not matter.
+				 */
+				WRITE_ONCE(mig_cpdc->is_stealer, false);
+				WRITE_ONCE(cpdc->is_stealee, false);
+				*sticky_cpdom = mig_cpdom;
+				break;
+			}
+		}
+	}
+
+	return cpu;
+}
+
+__hidden __noinline
+s32 pick_idle_cpu(struct pick_ctx *ctx, bool *is_idle)
 {
 	const struct cpumask *idle_cpumask = NULL, *idle_smtmask = NULL;
-	struct cpdom_ctx *cpdc, *mig_cpdc;
 	s32 cpu = -ENOENT, sticky_cpu;
+	s64 sticky_cpdom = -ENOENT;
+	struct cpdom_ctx *cpdc;
 	bool i_smt_empty;
-	s64 sticky_cpdom = -ENOENT, mig_cpdom, nr_nbr;
-	int i, j;
 
 	/*
 	 * At the high level, the idle CPU selection policy considers the
 	 * following factors:
-	 * 
+	 *
 	 * 1) Current active and overflow set: Stay on the current active and
 	 *    overflow sets if a task can run on them.
-	 * 
+	 *
 	 * 2) CPU preference order: If a task cannot run on the current active
 	 *    or overflow set, extend the overflow set following the CPU
 	 *    preference order (performance mode vs. power-save mode).
-	 * 
+	 *
 	 * 3) CPU type vs. task type: If possible, try to run a task on the
 	 *    matching CPU type (i.e., a big task on a big core vs. a little
 	 *    task on a little core). If the matching CPUs are not active,
 	 *    stay on the previous CPU.
-	 * 
+	 *
 	 * 4) Fully idle CPU vs. partially idle CPU: Choose a fully idle CPU
 	 *    over a partially idle CPU within the previous CPU's domain.
 	 *
@@ -581,7 +599,7 @@ s32 __attribute__ ((noinline)) pick_idle_cpu(struct pick_ctx *ctx, bool *is_idle
 	 *    good for cache locality because the waker task hands over the CPU
 	 *    to the wakee task for the further processing after finishing
 	 *    its job.
-	 * 
+	 *
 	 * 6) Minimize cross-domain migration: Before migrating to a neighbor
 	 *    domain, try to find an (any) idle CPU on the current domain.
 	 *    Migrate a task to another domain only when the current sticky
@@ -739,61 +757,22 @@ s32 __attribute__ ((noinline)) pick_idle_cpu(struct pick_ctx *ctx, bool *is_idle
 	 *  2) The sticky CPU or waker CPU is not idle.
 	 *  3) But there is at least one idle CPU in active/overflow set.
 	 *
+	 * Now, let’s consider migrating the task to the sticky domain’s
+	 * neighbor. If there is only one domain, let’s stay at the sticky
+	 * domain.
+	 * 
 	 * If there is a fully idle core in the system (i.e., !is_smt_empty),
-	 * let's migrate a task to another domain when
-	 *  1) The sticky domain is over-loaded (cpdc->is_stealee)
-	 *  2) The target domain is under-loaded (mig_cpdc->is_stealer)
-	 *     that has a fully idle core.
-	 *
-	 * Note that when a system is under-loaded, task donation works better
-	 * than task stealing because DSQs are mostly empty (i.e., it is hard
-	 * to steal from a DSQ).
+	 * let's try to migrate a task to another domain.
 	 */
-	if (i_smt_empty || nr_cpdoms == 1)
-		goto skip_fully_idle_neighbor;
-
-	cpdc = MEMBER_VPTR(cpdom_ctxs, [sticky_cpdom]);
-	if (!cpdc || !READ_ONCE(cpdc->is_stealee))
-		goto skip_fully_idle_neighbor;
-
-	mig_cpdom = sticky_cpdom;
-	bpf_for(i, 0, LAVD_CPDOM_MAX_DIST) {
-		nr_nbr = min(cpdc->nr_neighbors[i], LAVD_CPDOM_MAX_NR);
-		if (nr_nbr == 0)
-			break;
-		bpf_for(j, 0, LAVD_CPDOM_MAX_NR) {
-			if (j >= nr_nbr)
-				break;
-			mig_cpdom = get_neighbor_id(cpdc, i, j);
-			if (mig_cpdom < 0)
-				continue;
-
-			mig_cpdc = MEMBER_VPTR(cpdom_ctxs, [mig_cpdom]);
-			if (!mig_cpdc || !READ_ONCE(mig_cpdc->is_stealer))
-				continue;
-
-			cpu = pick_idle_cpu_at_cpdom(ctx, mig_cpdom, SCX_PICK_IDLE_CORE, is_idle);
-			if (cpu >= 0) {
-				/*
-				 * If task donation is successful, mark the stealer
-				 * and the stealee's job done. By marking done,
-				 * those compute domains would not be involved in
-				 * load balancing until the end of this round,
-				 * so this helps gradual migration. It is racy
-				 * in task stealings and donations, but we don't
-				 * care because a slight over-migration does not matter.
-				 */
-				WRITE_ONCE(mig_cpdc->is_stealer, false);
-				WRITE_ONCE(cpdc->is_stealee, false);
-				sticky_cpdom = mig_cpdom;
-				goto unlock_out;
-			}
-		}
+	if (!i_smt_empty && (nr_cpdoms > 1) &&
+	    (cpdc = MEMBER_VPTR(cpdom_ctxs, [sticky_cpdom])) &&
+	    READ_ONCE(cpdc->is_stealee)) {
+		cpu = migrate_to_neighbor(ctx, cpdc, SCX_PICK_IDLE_CORE,
+					  &sticky_cpdom, is_idle);
+		if (cpu >= 0)
+			goto unlock_out;
 	}
-skip_fully_idle_neighbor:
-	/* NOTE: There is no fully idle CPU in the neighboring domain,
-	 * it is not worth migrating. So try to find any idle CPU from
-	 * the sticky domain. */
+	/* NOTE: There is no fully idle CPU in the neighboring domain. */
 
 	/*
 	 * If there is an (partially) idle CPU in the sticky domain, stay on it.
@@ -804,6 +783,20 @@ skip_fully_idle_neighbor:
 		goto unlock_out;
 	/* NOTE: There is no even partially idle CPU in the sticky domain. */
 
+	/*
+	 * If aggressive migration is preferred for the task (e.g., freshly
+	 * execv()-ed task), find a partially idle CPU from the sticky
+	 * domain’s neighbor.
+	 */
+	if ((nr_cpdoms > 1) &&
+	    test_task_flag(ctx->taskc, LAVD_FLAG_MIGRATION_AGGRESSIVE) &&
+	    (cpdc = MEMBER_VPTR(cpdom_ctxs, [sticky_cpdom])) &&
+	    READ_ONCE(cpdc->is_stealee)) {
+		cpu = migrate_to_neighbor(ctx, cpdc, 0, &sticky_cpdom, is_idle);
+		if (cpu >= 0)
+			goto unlock_out;
+	}
+
 	/*
 	 * Instead of chasing a partially idle CPU in neighboring domains,
 	 * let's stay on the previous CPU or the sticky domain for cache
