@@ -75,7 +75,7 @@ const SCHEDULER_NAME: &str = "scx_lavd";
 /// See the more detailed overview of the LAVD design at main.bpf.c.
 #[derive(Debug, Parser)]
 struct Opts {
-    /// Depricated, noop, use RUST_LOG or --log-level instead.
+    /// Deprecated, noop, use RUST_LOG or --log-level instead.
     #[clap(short = 'v', long, action = clap::ArgAction::Count)]
     verbose: u8,
 
@@ -438,6 +438,12 @@ impl<'a> Scheduler<'a> {
         Self::init_cpus(&mut skel, &order);
         Self::init_cpdoms(&mut skel, &order);
 
+        // When there are multiple domains, hook the execve() syscall family
+        // to enable aggressive cross-domain migration when execve() is called.
+        if order.cpdom_map.len() > 1 {
+            Self::attach_execve_tracepoints(&mut skel)?;
+        }
+
         // Initialize skel according to @opts.
         Self::init_globals(&mut skel, &opts, &order, debug_level);
 
@@ -517,6 +523,21 @@ impl<'a> Scheduler<'a> {
         compat::cond_tracepoints_enable(tracepoints)
     }
 
+    fn attach_execve_tracepoints(skel: &mut OpenBpfSkel) -> Result<bool> {
+        let tracepoints = vec![
+            (
+                "syscalls:sys_enter_execve",
+                &skel.progs.cond_hook_sys_enter_execve,
+            ),
+            (
+                "syscalls:sys_enter_execveat",
+                &skel.progs.cond_hook_sys_enter_execveat,
+            ),
+        ];
+
+        compat::cond_tracepoints_enable(tracepoints)
+    }
+
     fn init_cpus(skel: &mut OpenBpfSkel, order: &CpuOrder) {
         debug!("{:#?}", order);
 
@@ -965,7 +986,7 @@ fn main(mut opts: Opts) -> Result<()> {
     init_log(&opts);
 
     if opts.verbose > 0 {
-        warn!("Setting verbose via -v is depricated and will be an error in future releases.");
+        warn!("Setting verbose via -v is deprecated and will be an error in future releases.");
     }
 
     if let Some(run_id) = opts.run_id {
