@@ -58,7 +58,7 @@ const NR_CSTATS: usize = bpf_intf::cell_stat_idx_NR_CSTATS as usize;
 /// split and which CPUs they should be assigned to.
 #[derive(Debug, Parser)]
 struct Opts {
-    /// Depricated, noop, use RUST_LOG or --log-level instead.
+    /// Deprecated, noop, use RUST_LOG or --log-level instead.
     #[clap(short = 'v', long, action = clap::ArgAction::Count)]
     verbose: u8,
 
@@ -113,6 +113,16 @@ struct Opts {
     #[clap(long, action = clap::ArgAction::SetTrue)]
     split_vtime_updates: bool,
 
+    /// Disable SCX cgroup callbacks (for when CPU cgroup controller is disabled).
+    /// Uses tracepoints and cgroup iteration instead.
+    #[clap(long, action = clap::ArgAction::SetTrue)]
+    cpu_controller_disabled: bool,
+
+    /// Reject tasks with multi-CPU pinning that doesn't cover the entire cell.
+    /// By default, these tasks are allowed but may have degraded performance.
+    #[clap(long, action = clap::ArgAction::SetTrue)]
+    reject_multicpu_pinning: bool,
+
     #[clap(flatten, next_help_heading = "Libbpf Options")]
     pub libbpf: LibbpfOpts,
 }
@@ -212,12 +222,23 @@ impl<'a> Scheduler<'a> {
             .unwrap()
             .exiting_task_workaround_enabled = opts.exiting_task_workaround;
         skel.maps.rodata_data.as_mut().unwrap().split_vtime_updates = opts.split_vtime_updates;
+        skel.maps
+            .rodata_data
+            .as_mut()
+            .unwrap()
+            .cpu_controller_disabled = opts.cpu_controller_disabled;
 
         skel.maps.rodata_data.as_mut().unwrap().nr_possible_cpus = *NR_CPUS_POSSIBLE as u32;
         for cpu in topology.all_cpus.keys() {
             skel.maps.rodata_data.as_mut().unwrap().all_cpus[cpu / 8] |= 1 << (cpu % 8);
         }
 
+        skel.maps
+            .rodata_data
+            .as_mut()
+            .unwrap()
+            .reject_multicpu_pinning = opts.reject_multicpu_pinning;
+
         match *compat::SCX_OPS_ALLOW_QUEUED_WAKEUP {
             0 => info!("Kernel does not support queued wakeup optimization."),
             v => skel.struct_ops.mitosis_mut().flags |= v,
@@ -597,7 +618,7 @@ fn main(opts: Opts) -> Result<()> {
     }
 
     if opts.verbose > 0 {
-        warn!("Setting verbose via -v is depricated and will be an error in future releases.");
+        warn!("Setting verbose via -v is deprecated and will be an error in future releases.");
     }
 
     debug!("opts={:?}", &opts);
