@@ -84,6 +84,15 @@ enum stat_idx {
 	P2DQ_STAT_WAKE_PREV,
 	P2DQ_STAT_WAKE_LLC,
 	P2DQ_STAT_WAKE_MIG,
+	P2DQ_STAT_FORK_BALANCE,
+	P2DQ_STAT_EXEC_BALANCE,
+	P2DQ_STAT_FORK_SAME_LLC,
+	P2DQ_STAT_EXEC_SAME_LLC,
+	P2DQ_STAT_THERMAL_KICK,
+	P2DQ_STAT_THERMAL_AVOID,
+	P2DQ_STAT_EAS_LITTLE_SELECT,
+	P2DQ_STAT_EAS_BIG_SELECT,
+	P2DQ_STAT_EAS_FALLBACK,
 	P2DQ_NR_STATS,
 };
 
