@@ -10,7 +10,7 @@
 #ifdef LSP
 #define __bpf__
 #include "../../../../include/scx/common.bpf.h"
-#include "../../../../include/scx/bpf_arena_common.bpf.h"
+#include "../../../../include/bpf_arena_common.bpf.h"
 #include "../../../../include/scx/percpu.bpf.h"
 #include "../../../../include/lib/atq.h"
 #include "../../../../include/lib/cpumask.h"
@@ -21,7 +21,7 @@
 #include "../../../../include/lib/topology.h"
 #else
 #include <scx/common.bpf.h>
-#include <scx/bpf_arena_common.bpf.h>
+#include <bpf_arena_common.bpf.h>
 #include <scx/percpu.bpf.h>
 #include <lib/atq.h>
 #include <lib/cpumask.h>
@@ -50,6 +50,11 @@ char _license[] SEC("license") = "GPL";
 
 UEI_DEFINE(uei);
 
+
+#ifndef PF_FORKNOEXEC
+#define PF_FORKNOEXEC 0x00000040
+#endif
+
 #define dbg(fmt, args...)	do { if (debug) bpf_printk(fmt, ##args); } while (0)
 #define trace(fmt, args...)	do { if (debug > 1) bpf_printk(fmt, ##args); } while (0)
 
@@ -130,6 +135,12 @@ const volatile struct {
 	bool keep_running_enabled;
 	bool kthreads_local;
 	bool pelt_enabled;
+	bool fork_balance;
+	bool exec_balance;
+	bool enable_eas;
+	bool thermal_enabled;
+	u16 small_task_threshold;
+	u16 large_task_threshold;
 } p2dq_config = {
 	.sched_mode = MODE_DEFAULT,
 	.nr_dsqs_per_llc = 3,
@@ -149,6 +160,12 @@ const volatile struct {
 	.keep_running_enabled = true,
 	.kthreads_local = true,
 	.pelt_enabled = true,
+	.fork_balance = true,
+	.exec_balance = true,
+	.enable_eas = false,
+	.thermal_enabled = false,
+	.small_task_threshold = 256,
+	.large_task_threshold = 768,
 };
 
 /* Latency priority and preemption configuration */
@@ -178,6 +195,10 @@ u64 cpu_node_ids[MAX_CPUS];
 u64 big_core_ids[MAX_CPUS];
 u64 dsq_time_slices[MAX_DSQS_PER_LLC];
 
+/* Energy and capacity per CPU for energy-aware scheduling */
+u16 cpu_energy_cost[MAX_CPUS];  // Energy cost coefficient (0-65535)
+u16 cpu_capacity[MAX_CPUS];     // CPU capacity (0-1024)
+
 /* DHQ per LLC pair for migration (MAX_LLCS / 2 DHQs) */
 scx_dhq_t *llc_pair_dhqs[MAX_LLCS / 2];
 /* Track number of LLCs per NUMA node for strand assignment */
@@ -323,6 +344,13 @@ static __always_inline u32 pelt_decay(u32 val, u32 periods)
 	return val;
 }
 
+/* Forward declarations for energy-aware scheduling helpers */
+static __always_inline u32 get_cpu_capacity(s32 cpu);
+static __always_inline u32 get_cpu_energy_cost(s32 cpu);
+static __always_inline u32 get_task_util(struct task_struct *p);
+static __always_inline bool prefer_little_core(struct task_struct *p);
+static __always_inline bool prefer_big_core(struct task_struct *p);
+
 /*
  * Update task's PELT metrics based on runtime.
  * Called when task stops running or starts running (for decay).
@@ -330,11 +358,14 @@ static __always_inline u32 pelt_decay(u32 val, u32 periods)
  * @taskc: Task context to update
  * @now: Current timestamp in ns
  * @delta_ns: Runtime delta (0 for decay-only update)
+ * @task_cpu: CPU the task is running on
  */
-static __always_inline void update_task_pelt(task_ctx *taskc, u64 now, u64 delta_ns)
+static __always_inline void update_task_pelt(task_ctx *taskc, u64 now, u64 delta_ns, s32 task_cpu)
 {
 	u64 elapsed_ns, elapsed_ms;
 	u32 periods, delta_ms;
+	u32 capacity, freq;
+	u64 scaled_delta_ms, scaled_period_contrib;
 
 	if (!p2dq_config.pelt_enabled)
 		return;
@@ -369,13 +400,27 @@ static __always_inline void update_task_pelt(task_ctx *taskc, u64 now, u64 delta
 		taskc->util_sum = pelt_decay(taskc->util_sum, periods);
 	}
 
+	capacity = get_cpu_capacity(task_cpu);
+	freq = scx_bpf_cpuperf_cur(task_cpu);
+	if (freq == 0)
+		freq = SCX_CPUPERF_ONE;
+
+	/*
+	 * Scale period contribution by capacity and frequency
+	 * This makes the PELT metric represent "work done at max CPU capacity at max freq"
+	 *
+	 * Formula: scaled_time = wall_time * (capacity / 1024) * (freq / 1024)
+	 *         = wall_time * capacity * freq / (1024 * 1024)
+	 */
 	if (taskc->period_contrib > 0) {
-		taskc->util_sum += taskc->period_contrib;
+		scaled_period_contrib = (taskc->period_contrib * capacity * freq) / (1024ULL * 1024ULL);
+		taskc->util_sum += scaled_period_contrib;
 		taskc->period_contrib = 0;
 	}
 
 	delta_ms = delta_ns / NSEC_PER_MSEC;
-	taskc->util_sum += delta_ms;
+	scaled_delta_ms = (delta_ms * capacity * freq) / (1024ULL * 1024ULL);
+	taskc->util_sum += scaled_delta_ms;
 
 	if (unlikely(taskc->util_sum > PELT_SUM_MAX))
 		taskc->util_sum = PELT_SUM_MAX;
@@ -683,6 +728,125 @@ static task_ctx *lookup_task_ctx(struct task_struct *p)
 	return taskc;
 }
 
+/*
+ * Get CPU capacity (0-1024)
+ * Used for frequency/capacity-invariant PELT and energy-aware scheduling
+ */
+static __always_inline u32 get_cpu_capacity(s32 cpu)
+{
+	if (cpu < 0 || cpu >= MAX_CPUS)
+		return 1024;
+	return cpu_capacity[cpu] ? cpu_capacity[cpu] : 1024;
+}
+
+/*
+ * Get CPU energy cost (lower is more efficient)
+ */
+static __always_inline u32 get_cpu_energy_cost(s32 cpu)
+{
+	if (cpu < 0 || cpu >= MAX_CPUS)
+		return 65535;
+	return cpu_energy_cost[cpu] ? cpu_energy_cost[cpu] : 1024;
+}
+
+/*
+ * Get task utilization from custom PELT
+ * Returns util_avg in range 0-1024
+ * NOTE: Frequency and capacity invariant after modifications
+ */
+static __always_inline u32 get_task_util(struct task_struct *p)
+{
+	task_ctx *taskc;
+
+	taskc = lookup_task_ctx(p);
+	if (!taskc)
+		return 0;
+
+	return taskc->util_avg;
+}
+
+/*
+ * Check if task should prefer little cores based on utilization
+ */
+static __always_inline bool prefer_little_core(struct task_struct *p)
+{
+	if (!p2dq_config.enable_eas || !topo_config.has_little_cores)
+		return false;
+
+	u32 util = get_task_util(p);
+	return util < p2dq_config.small_task_threshold;
+}
+
+/*
+ * Check if task should prefer big cores based on utilization
+ */
+static __always_inline bool prefer_big_core(struct task_struct *p)
+{
+	if (!p2dq_config.enable_eas || !topo_config.has_little_cores)
+		return false;
+
+	u32 util = get_task_util(p);
+	return util > p2dq_config.large_task_threshold;
+}
+
+/*
+ * Get effective CPU capacity accounting for thermal pressure AND frequency
+ * Returns capacity in range 0-1024
+ */
+static __always_inline u32 get_effective_cpu_capacity(s32 cpu)
+{
+	struct cpu_ctx *cpuc;
+	u32 base_capacity, thermal_pressure;
+	u32 cur_freq;
+	u64 effective_capacity;
+
+	if (cpu < 0 || cpu >= MAX_CPUS)
+		return 0;
+
+	cpuc = lookup_cpu_ctx(cpu);
+	if (!cpuc)
+		return 0;
+
+	base_capacity = get_cpu_capacity(cpu);
+
+	thermal_pressure = cpuc->perf;
+
+	cur_freq = scx_bpf_cpuperf_cur(cpu);
+	if (cur_freq == 0)
+		cur_freq = SCX_CPUPERF_ONE;
+
+	/*
+	 * Effective capacity = (base - thermal) * freq / 1024
+	 * Combines thermal throttling and frequency scaling
+	 */
+	if (thermal_pressure >= base_capacity)
+		return 0;  /* Fully throttled */
+
+	effective_capacity = (u64)(base_capacity - thermal_pressure) * cur_freq / SCX_CPUPERF_ONE;
+
+	return (u32)effective_capacity;
+}
+
+/*
+ * Check if CPU is thermally throttled
+ * Returns true if pressure > 25% capacity loss
+ */
+static __always_inline bool is_cpu_throttled(s32 cpu)
+{
+	struct cpu_ctx *cpuc;
+
+	if (cpu < 0 || cpu >= MAX_CPUS)
+		return false;
+
+	cpuc = lookup_cpu_ctx(cpu);
+	if (!cpuc)
+		return false;
+
+	/* Throttled if pressure > 256 (25% of 1024) */
+	return cpuc->perf > 256;
+}
+
+
 struct {
 	__uint(type, BPF_MAP_TYPE_PERCPU_ARRAY);
 	__type(key, u32);
@@ -825,7 +989,7 @@ static bool keep_running(struct cpu_ctx *cpuc, struct llc_ctx *llcx,
 static s32 pick_idle_affinitized_cpu(struct task_struct *p, task_ctx *taskc,
 				     s32 prev_cpu, bool *is_idle)
 {
-	const struct cpumask *idle_smtmask, *idle_cpumask;
+	const struct cpumask *idle_cpumask;
 	struct mask_wrapper *wrapper;
 	struct bpf_cpumask *mask;
 	struct llc_ctx *llcx;
@@ -838,7 +1002,6 @@ static s32 pick_idle_affinitized_cpu(struct task_struct *p, task_ctx *taskc,
 	}
 
 	idle_cpumask = scx_bpf_get_idle_cpumask();
-	idle_smtmask = scx_bpf_get_idle_smtmask();
 
 	if (!(llcx = lookup_llc_ctx(taskc->llc_id)) ||
 	    !llcx->cpumask)
@@ -892,28 +1055,320 @@ static s32 pick_idle_affinitized_cpu(struct task_struct *p, task_ctx *taskc,
 		}
 	}
 
-	// Fallback to anywhere the task can run
 	cpu = bpf_cpumask_any_distribute(p->cpus_ptr);
 
 found_cpu:
 	scx_bpf_put_cpumask(idle_cpumask);
-	scx_bpf_put_cpumask(idle_smtmask);
 
 	return cpu;
 }
 
+/**
+ * find_idle_cpu_in_target_llc - Find idle CPU in specific LLC for fork/exec balancing
+ * @p: task to place
+ * @target_llc_id: LLC to search for idle CPU
+ *
+ * Returns CPU ID of idle CPU in target LLC, or -1 if none available.
+ */
+static __always_inline s32 find_idle_cpu_in_target_llc(struct task_struct *p, u32 target_llc_id)
+{
+	const struct cpumask *idle_smtmask, *idle_cpumask;
+	struct llc_ctx *llcx;
+	s32 cpu = -1;
+
+	if (!p || target_llc_id >= MAX_LLCS)
+		return -1;
+
+	llcx = lookup_llc_ctx(target_llc_id);
+	if (!llcx)
+		return -1;
+
+	idle_cpumask = scx_bpf_get_idle_cpumask();
+	idle_smtmask = scx_bpf_get_idle_smtmask();
+	if (!idle_cpumask || !idle_smtmask)
+		goto out;
+
+	/* Try idle core first (both SMT siblings idle) */
+	bpf_for(cpu, 0, topo_config.nr_cpus) {
+		struct cpu_ctx *cpuc = lookup_cpu_ctx(cpu);
+		if (!cpuc || cpuc->llc_id != target_llc_id)
+			continue;
+		if (!bpf_cpumask_test_cpu(cpu, p->cpus_ptr))
+			continue;
+		if (bpf_cpumask_test_cpu(cpu, idle_smtmask) &&
+		    scx_bpf_test_and_clear_cpu_idle(cpu))
+			goto out;
+	}
+
+	/* No idle core, try any idle CPU */
+	bpf_for(cpu, 0, topo_config.nr_cpus) {
+		struct cpu_ctx *cpuc = lookup_cpu_ctx(cpu);
+		if (!cpuc || cpuc->llc_id != target_llc_id)
+			continue;
+		if (!bpf_cpumask_test_cpu(cpu, p->cpus_ptr))
+			continue;
+		if (bpf_cpumask_test_cpu(cpu, idle_cpumask) &&
+		    scx_bpf_test_and_clear_cpu_idle(cpu))
+			goto out;
+	}
+
+	cpu = -1;
+
+out:
+	scx_bpf_put_cpumask(idle_cpumask);
+	scx_bpf_put_cpumask(idle_smtmask);
+	return cpu;
+}
+
+/**
+ * find_least_loaded_llc_for_fork - Find least loaded LLC using pick-2
+ * @parent_llc_id: LLC ID of parent task
+ *
+ * Returns LLC ID of least loaded LLC.
+ * Made noinline to reduce verifier complexity (only scalar params).
+ */
+u32 __attribute__((noinline)) find_least_loaded_llc_for_fork(u32 parent_llc_id)
+{
+	struct llc_ctx *parent_llc, *candidate_llc;
+	u32 candidate_id, best_id;
+	u64 best_load;
+
+	if (parent_llc_id >= MAX_LLCS)
+		return parent_llc_id;
+
+	parent_llc = lookup_llc_ctx(parent_llc_id);
+	if (!parent_llc)
+		return parent_llc_id;
+
+	best_id = parent_llc_id;
+	best_load = parent_llc->load;
+
+	if (topo_config.nr_llcs == 2) {
+		candidate_id = (parent_llc_id == llc_ids[0]) ? llc_ids[1] : llc_ids[0];
+		candidate_llc = lookup_llc_ctx(candidate_id);
+		if (candidate_llc && candidate_llc->load <= best_load)
+			return candidate_id;
+		return best_id;
+	}
+
+	candidate_llc = rand_llc_ctx();
+	if (candidate_llc && candidate_llc->load <= best_load)
+		return candidate_llc->id;
+
+	return best_id;
+}
+
+/*
+ * Pick idle CPU from mask, avoiding thermally throttled CPUs.
+ * Simpler/faster than full energy-aware selection - used for MODE_PERF/MODE_EFFICIENCY.
+ */
+static __always_inline s32 pick_idle_thermal_aware(struct bpf_cpumask *mask,
+						   struct task_struct *p)
+{
+	s32 cpu, best_cpu = -1;
+	u32 best_capacity = 0;
+
+	if (!mask)
+		return -1;
+
+	/* First pass: try to find unthrottled idle CPU */
+	bpf_for(cpu, 0, topo_config.nr_cpus) {
+		if (!bpf_cpumask_test_cpu(cpu, cast_mask(mask)))
+			continue;
+		if (!bpf_cpumask_test_cpu(cpu, p->cpus_ptr))
+			continue;
+		if (is_cpu_throttled(cpu))
+			continue;
+		if (scx_bpf_test_and_clear_cpu_idle(cpu))
+			return cpu;
+	}
+
+	/* Second pass: allow throttled CPUs, prefer least throttled */
+	bpf_for(cpu, 0, topo_config.nr_cpus) {
+		u32 capacity;
+
+		if (mask && !bpf_cpumask_test_cpu(cpu, cast_mask(mask)))
+			continue;
+		if (!bpf_cpumask_test_cpu(cpu, p->cpus_ptr))
+			continue;
+		if (!scx_bpf_test_and_clear_cpu_idle(cpu))
+			continue;
+
+		capacity = get_effective_cpu_capacity(cpu);
+		if (capacity > best_capacity) {
+			best_capacity = capacity;
+			best_cpu = cpu;
+		}
+	}
+
+	if (best_cpu >= 0)
+		stat_inc(P2DQ_STAT_THERMAL_AVOID);
+
+	return best_cpu;
+}
+
+/*
+ * Select best idle CPU from mask based on:
+ * - Not thermally throttled (priority 1)
+ * - High effective capacity (accounts for thermal + freq)
+ * - Low energy cost
+ *
+ * Returns CPU ID or -1 if no suitable CPU found
+ * Updates best_score with score of selected CPU (higher is better)
+ */
+static __always_inline s32 select_best_idle_cpu(struct task_struct *p,
+						struct bpf_cpumask *mask,
+						u32 *best_score)
+{
+	s32 cpu, best_cpu = -1;
+	u32 highest_score = 0;
+
+	if (!mask || !best_score)
+		return -1;
+
+	*best_score = 0;
+
+	bpf_for(cpu, 0, topo_config.nr_cpus) {
+		u32 capacity, energy_cost, score;
+
+		if (!bpf_cpumask_test_cpu(cpu, cast_mask(mask)))
+			continue;
+		if (!bpf_cpumask_test_cpu(cpu, p->cpus_ptr))
+			continue;
+		if (!scx_bpf_test_and_clear_cpu_idle(cpu))
+			continue;
+
+		capacity = get_effective_cpu_capacity(cpu);
+		energy_cost = get_cpu_energy_cost(cpu);
+
+		/*
+		 * Score formula: prioritize capacity, penalize energy cost
+		 * score = capacity * 10 - (energy_cost / 10)
+		 * Higher score is better
+		 *
+		 * Throttled CPUs (capacity=0) get score=0
+		 */
+		if (is_cpu_throttled(cpu)) {
+			score = 0;  /* Heavily penalize throttled CPUs */
+		} else {
+			/* Multiply capacity by 10 for more weight vs energy cost */
+			score = (capacity * 10);
+			/* Subtract scaled energy cost (divide by 10 to reduce impact) */
+			if (energy_cost < score)
+				score -= (energy_cost / 10);
+			else
+				score = 1;
+		}
+
+		if (score > highest_score) {
+			highest_score = score;
+			best_cpu = cpu;
+		}
+	}
+
+	*best_score = highest_score;
+	return best_cpu;
+}
+
+/*
+ * Pick idle CPU using comprehensive energy-aware scheduling
+ * Tries preferred core type first, then fallback type
+ */
+static __always_inline s32 pick_idle_energy_aware(struct task_struct *p,
+						  struct llc_ctx *llcx,
+						  bool *is_idle)
+{
+	s32 cpu = -1;
+	u32 pref_score = 0, fallback_score = 0;
+
+	if (!llcx || !is_idle)
+		return -1;
+
+	/* Determine preferred and fallback cpumasks based on task utilization */
+	struct bpf_cpumask *pref_mask = NULL;
+	struct bpf_cpumask *fallback_mask = NULL;
+	bool prefer_little = prefer_little_core(p);
+	bool prefer_big = prefer_big_core(p);
+
+	if (prefer_little) {
+		pref_mask = llcx->little_cpumask;
+		fallback_mask = llcx->big_cpumask;
+	} else if (prefer_big) {
+		pref_mask = llcx->big_cpumask;
+		fallback_mask = llcx->little_cpumask;
+	} else {
+		/* No strong preference, try both and pick best score */
+		s32 little_cpu = -1, big_cpu = -1;
+		u32 little_score = 0, big_score = 0;
+
+		if (llcx->little_cpumask)
+			little_cpu = select_best_idle_cpu(p, llcx->little_cpumask, &little_score);
+		if (llcx->big_cpumask)
+			big_cpu = select_best_idle_cpu(p, llcx->big_cpumask, &big_score);
+
+		/* Pick whichever has better score */
+		if (little_cpu >= 0 && big_cpu >= 0) {
+			if (little_score >= big_score) {
+				*is_idle = true;
+				stat_inc(P2DQ_STAT_EAS_LITTLE_SELECT);
+				return little_cpu;
+			} else {
+				*is_idle = true;
+				stat_inc(P2DQ_STAT_EAS_BIG_SELECT);
+				return big_cpu;
+			}
+		} else if (little_cpu >= 0) {
+			*is_idle = true;
+			stat_inc(P2DQ_STAT_EAS_LITTLE_SELECT);
+			return little_cpu;
+		} else if (big_cpu >= 0) {
+			*is_idle = true;
+			stat_inc(P2DQ_STAT_EAS_BIG_SELECT);
+			return big_cpu;
+		}
+		return -1;
+	}
+
+	/* Try preferred core type first */
+	if (pref_mask) {
+		cpu = select_best_idle_cpu(p, pref_mask, &pref_score);
+		if (cpu >= 0) {
+			*is_idle = true;
+			if (prefer_little)
+				stat_inc(P2DQ_STAT_EAS_LITTLE_SELECT);
+			else
+				stat_inc(P2DQ_STAT_EAS_BIG_SELECT);
+			return cpu;
+		}
+	}
+
+	/* Fallback to opposite core type if preferred not available */
+	if (fallback_mask) {
+		cpu = select_best_idle_cpu(p, fallback_mask, &fallback_score);
+		if (cpu >= 0) {
+			*is_idle = true;
+			stat_inc(P2DQ_STAT_EAS_FALLBACK);
+			if (prefer_little)
+				stat_inc(P2DQ_STAT_EAS_BIG_SELECT);
+			else
+				stat_inc(P2DQ_STAT_EAS_LITTLE_SELECT);
+			return cpu;
+		}
+	}
+
+	return -1;
+}
+
 static s32 pick_idle_cpu(struct task_struct *p, task_ctx *taskc,
 			 s32 prev_cpu, u64 wake_flags, bool *is_idle)
 {
-	const struct cpumask *idle_smtmask, *idle_cpumask;
+	const struct cpumask *idle_cpumask;
 	struct llc_ctx *llcx;
 	s32 pref_cpu, cpu = prev_cpu;
 	bool migratable = false;
 
 	idle_cpumask = scx_bpf_get_idle_cpumask();
-	idle_smtmask = scx_bpf_get_idle_smtmask();
-
-	if (!idle_cpumask || !idle_smtmask)
+	if (!idle_cpumask)
 		goto found_cpu;
 
 	if (bpf_cpumask_test_cpu(prev_cpu, idle_cpumask) &&
@@ -1037,39 +1492,66 @@ static s32 pick_idle_cpu(struct task_struct *p, task_ctx *taskc,
 		goto found_cpu;
 	}
 
+	/*
+	 * Energy-aware selection with comprehensive scoring
+	 * Uses effective capacity, energy cost, and thermal awareness
+	 */
+	if (p2dq_config.enable_eas && topo_config.has_little_cores) {
+		cpu = pick_idle_energy_aware(p, llcx, is_idle);
+		if (cpu >= 0)
+			goto found_cpu;
+	}
+
 	if (p2dq_config.sched_mode == MODE_PERF &&
 	    topo_config.has_little_cores &&
 	    llcx->big_cpumask) {
-		cpu = __pick_idle_cpu(llcx->big_cpumask,
-				      SCX_PICK_IDLE_CORE);
-		if (cpu >= 0) {
-			*is_idle = true;
-			goto found_cpu;
-		}
-		if (llcx->big_cpumask) {
-			cpu = __pick_idle_cpu(llcx->big_cpumask, 0);
+		/* Try thermal-aware selection first for big cores if thermal tracking enabled */
+		if (p2dq_config.thermal_enabled) {
+			cpu = pick_idle_thermal_aware(llcx->big_cpumask, p);
 			if (cpu >= 0) {
 				*is_idle = true;
 				goto found_cpu;
 			}
 		}
+		/* Fallback to non-thermal-aware if thermal disabled or no idle big cores */
+		if (llcx->big_cpumask &&
+		    (cpu = __pick_idle_cpu(llcx->big_cpumask, SCX_PICK_IDLE_CORE)) &&
+		    cpu >= 0) {
+			*is_idle = true;
+			goto found_cpu;
+		}
+		if (llcx->big_cpumask &&
+		    (cpu = __pick_idle_cpu(llcx->big_cpumask, 0)) &&
+		    cpu >= 0) {
+			*is_idle = true;
+			goto found_cpu;
+		}
 	}
 
 	if (p2dq_config.sched_mode == MODE_EFFICIENCY &&
 	    topo_config.has_little_cores &&
 	    llcx->little_cpumask) {
-		cpu = __pick_idle_cpu(llcx->little_cpumask, SCX_PICK_IDLE_CORE);
-		if (cpu >= 0) {
-			*is_idle = true;
-			goto found_cpu;
-		}
-		if (llcx->little_cpumask) {
-			cpu = __pick_idle_cpu(llcx->little_cpumask, 0);
+		/* Try thermal-aware selection first for little cores if thermal tracking enabled */
+		if (p2dq_config.thermal_enabled) {
+			cpu = pick_idle_thermal_aware(llcx->little_cpumask, p);
 			if (cpu >= 0) {
 				*is_idle = true;
 				goto found_cpu;
 			}
 		}
+		/* Fallback to non-thermal-aware if thermal disabled or no idle little cores */
+		if (llcx->little_cpumask &&
+		    (cpu = __pick_idle_cpu(llcx->little_cpumask, SCX_PICK_IDLE_CORE)) &&
+		    cpu >= 0) {
+			*is_idle = true;
+			goto found_cpu;
+		}
+		if (llcx->little_cpumask &&
+		    (cpu = __pick_idle_cpu(llcx->little_cpumask, 0)) &&
+		    cpu >= 0) {
+			*is_idle = true;
+			goto found_cpu;
+		}
 	}
 
 
@@ -1082,25 +1564,6 @@ static s32 pick_idle_cpu(struct task_struct *p, task_ctx *taskc,
 		stat_inc(P2DQ_STAT_SELECT_PICK2);
 	}
 
-	if (topo_config.has_little_cores &&
-	    llcx->little_cpumask && llcx->big_cpumask) {
-		if (task_ctx_test_flag(taskc, TASK_CTX_F_INTERACTIVE)) {
-			cpu = __pick_idle_cpu(llcx->little_cpumask,
-					      0);
-			if (cpu >= 0) {
-				*is_idle = true;
-				goto found_cpu;
-			}
-		} else {
-			cpu = __pick_idle_cpu(llcx->big_cpumask,
-					      SCX_PICK_IDLE_CORE);
-			if (cpu >= 0) {
-				*is_idle = true;
-				goto found_cpu;
-			}
-		}
-	}
-
 	if (p2dq_config.cpu_priority) {
 		pref_cpu = pref_idle_cpu(llcx);
 		if (llcx->cpumask && pref_cpu >= 0 &&
@@ -1112,6 +1575,21 @@ static s32 pick_idle_cpu(struct task_struct *p, task_ctx *taskc,
 		}
 	}
 
+	if (topo_config.has_little_cores && llcx->big_cpumask) {
+		cpu = __pick_idle_cpu(llcx->big_cpumask,
+				      SCX_PICK_IDLE_CORE);
+		if (cpu >= 0) {
+			*is_idle = true;
+			goto found_cpu;
+		}
+		if (llcx->little_cpumask &&
+		    (cpu = __pick_idle_cpu(llcx->little_cpumask,
+					   SCX_PICK_IDLE_CORE)) >= 0) {
+			*is_idle = true;
+			goto found_cpu;
+		}
+	}
+
 	// Next try in the local LLC (usually succeeds)
 	if (likely(llcx->cpumask &&
 	    (cpu = __pick_idle_cpu(llcx->cpumask,
@@ -1166,7 +1644,6 @@ static s32 pick_idle_cpu(struct task_struct *p, task_ctx *taskc,
 
 found_cpu:
 	scx_bpf_put_cpumask(idle_cpumask);
-	scx_bpf_put_cpumask(idle_smtmask);
 
 	return cpu;
 }
@@ -1240,7 +1717,7 @@ skip_preempt:
  * immediately in `enqueue` or later in `dispatch`. This returns a tagged union
  * with three states:
  * - P2DQ_ENQUEUE_PROMISE_COMPLETE: The enqueue has been completed. Note that
- *     this case _must_ be determinstic, or else scx_chaos will stall. That is,
+ *     this case _must_ be deterministic, or else scx_chaos will stall. That is,
  *     if the same task and enq_flags arrive twice, it must have returned
  *     _COMPLETE the first time to return it again.
  * - P2DQ_ENQUEUE_PROMISE_FIFO: The completer should enqueue this task on a fifo dsq.
@@ -1282,6 +1759,77 @@ static void async_p2dq_enqueue(struct enqueue_promise *ret,
 		return;
 	}
 
+	/* Exec balancing: balance tasks transitioning from fork to exec */
+	if (task_ctx_test_flag(taskc, TASK_CTX_F_FORKNOEXEC) && !(p->flags & PF_FORKNOEXEC) &&
+	    p2dq_config.exec_balance &&
+	    !lb_config.single_llc_mode &&
+	    task_ctx_test_flag(taskc, TASK_CTX_F_ALL_CPUS)) {
+		struct cpu_ctx *curr_cpuc = lookup_cpu_ctx(cpu);
+		if (curr_cpuc) {
+			u32 target_llc = find_least_loaded_llc_for_fork(curr_cpuc->llc_id);
+
+			if (target_llc != curr_cpuc->llc_id && target_llc < MAX_LLCS) {
+				taskc->target_llc_hint = target_llc;
+				stat_inc(P2DQ_STAT_EXEC_BALANCE);
+			} else {
+				stat_inc(P2DQ_STAT_EXEC_SAME_LLC);
+			}
+		}
+	}
+
+	if (p->flags & PF_FORKNOEXEC)
+		task_ctx_set_flag(taskc, TASK_CTX_F_FORKNOEXEC);
+	else
+		task_ctx_clear_flag(taskc, TASK_CTX_F_FORKNOEXEC);
+
+	/* Fork balancing: balance newly forked tasks across LLCs */
+	if (task_ctx_test_flag(taskc, TASK_CTX_F_FORKNOEXEC) && taskc->llc_runs == 0 &&
+	    p2dq_config.fork_balance &&
+	    !lb_config.single_llc_mode &&
+	    task_ctx_test_flag(taskc, TASK_CTX_F_ALL_CPUS)) {
+		struct cpu_ctx *curr_cpuc = lookup_cpu_ctx(cpu);
+		if (curr_cpuc) {
+			u32 target_llc = find_least_loaded_llc_for_fork(curr_cpuc->llc_id);
+
+			if (target_llc != curr_cpuc->llc_id && target_llc < MAX_LLCS) {
+				taskc->target_llc_hint = target_llc;
+				stat_inc(P2DQ_STAT_FORK_BALANCE);
+			} else {
+				stat_inc(P2DQ_STAT_FORK_SAME_LLC);
+			}
+		}
+	}
+
+	if (topo_config.nr_llcs > 1 &&
+	    taskc->target_llc_hint < MAX_LLCS) {
+		u32 target_llc_id = taskc->target_llc_hint;
+		s32 target_cpu;
+
+		taskc->target_llc_hint = MAX_LLCS;
+
+		target_cpu = find_idle_cpu_in_target_llc(p, target_llc_id);
+		if (target_cpu >= 0) {
+			struct cpu_ctx *target_cpuc = lookup_cpu_ctx(target_cpu);
+			struct llc_ctx *target_llc = lookup_llc_ctx(target_llc_id);
+
+			if (target_cpuc && target_llc) {
+				taskc->llc_id = target_llc_id;
+				taskc->llc_runs = 0;
+
+				update_vtime(p, target_cpuc, taskc, target_llc);
+				ret->kind = P2DQ_ENQUEUE_PROMISE_FIFO;
+				ret->cpu = target_cpu;
+				ret->fifo.dsq_id = SCX_DSQ_LOCAL_ON | target_cpu;
+				ret->fifo.slice_ns = taskc->slice_ns;
+				ret->fifo.enq_flags = enq_flags;
+
+				dbg("FORK/EXEC: pid=%d -> cpu=%d llc=%u",
+				    p->pid, target_cpu, target_llc_id);
+				return;
+			}
+		}
+	}
+
 	// Handle affinitized tasks: always use per-CPU affn_dsq
 	// All affinitized tasks queued to affn_dsq regardless of affinity breadth
 	if (!task_ctx_test_flag(taskc, TASK_CTX_F_ALL_CPUS) ||
@@ -1738,7 +2286,7 @@ static int p2dq_running_impl(struct task_struct *p)
 
 	/* Decay PELT metrics when task starts running (0 delta for decay-only) */
 	if (p2dq_config.pelt_enabled)
-		update_task_pelt(taskc, now, 0);
+		update_task_pelt(taskc, now, 0, task_cpu);
 
 	return 0;
 }
@@ -1750,6 +2298,7 @@ void BPF_STRUCT_OPS(p2dq_stopping, struct task_struct *p, bool runnable)
 	struct cpu_ctx *cpuc;
 	u64 used, scaled_used, last_dsq_slice_ns;
 	u64 now = bpf_ktime_get_ns();
+	s32 task_cpu = scx_bpf_task_cpu(p);
 
 	if (unlikely(!(taskc = lookup_task_ctx(p)) ||
 	    !(llcx = lookup_llc_ctx(taskc->llc_id))))
@@ -1762,11 +2311,8 @@ void BPF_STRUCT_OPS(p2dq_stopping, struct task_struct *p, bool runnable)
 		return;
 	}
 
-	// This is an optimization to not have to lookup the cpu_ctx every
-	// time. When a nice task was run we need to update the cpu_ctx so that
-	// tasks are no longer enqueued to the local DSQ.
 	if (task_ctx_test_flag(taskc, TASK_CTX_F_WAS_NICE) &&
-	    (cpuc = lookup_cpu_ctx(scx_bpf_task_cpu(p)))) {
+	    (cpuc = lookup_cpu_ctx(task_cpu))) {
 		cpu_ctx_clear_flag(cpuc, CPU_CTX_F_NICE_TASK);
 		task_ctx_clear_flag(taskc, TASK_CTX_F_WAS_NICE);
 	}
@@ -1784,14 +2330,12 @@ void BPF_STRUCT_OPS(p2dq_stopping, struct task_struct *p, bool runnable)
 
 	/* Update PELT metrics if enabled */
 	if (p2dq_config.pelt_enabled) {
-		update_task_pelt(taskc, now, used);
+		update_task_pelt(taskc, now, used, task_cpu);
 		aggregate_pelt_to_llc(llcx, taskc,
 				      task_ctx_test_flag(taskc, TASK_CTX_F_INTERACTIVE),
 				      !task_ctx_test_flag(taskc, TASK_CTX_F_ALL_CPUS));
-	}
-
-	/* Legacy load tracking (when PELT disabled) */
-	if (!p2dq_config.pelt_enabled) {
+	} else {
+		/* Legacy load tracking (when PELT disabled) */
 		__sync_fetch_and_add(&llcx->load, used);
 		if (taskc->dsq_index >= 0 && taskc->dsq_index < MAX_DSQS_PER_LLC)
 			__sync_fetch_and_add(&llcx->dsq_load[taskc->dsq_index], used);
@@ -1902,7 +2446,6 @@ static bool consume_llc(struct llc_ctx *llcx)
 			goto try_dsq;
 		}
 
-		/* Insert to LLC DSQ and let move_to_local handle affinity atomically */
 		trace("DHQ %llu insert %s[%d] to LLC DSQ",
 		      llcx->mig_dhq, p->comm, p->pid);
 		scx_bpf_dsq_insert_vtime(p,
@@ -1912,7 +2455,6 @@ static bool consume_llc(struct llc_ctx *llcx)
 					 taskc->enq_flags);
 		bpf_task_release(p);
 
-		/* Try to dispatch from LLC DSQ (handles affinity check atomically) */
 		if (scx_bpf_dsq_move_to_local(cpuc->llc_dsq))
 			return true;
 
@@ -1926,7 +2468,6 @@ static bool consume_llc(struct llc_ctx *llcx)
 			return false;
 		}
 
-/* Insert to LLC DSQ and let move_to_local handle affinity atomically */
 		trace("ATQ %llu insert %s[%d] to LLC DSQ",
 		      llcx->mig_atq, p->comm, p->pid);
 		scx_bpf_dsq_insert_vtime(p,
@@ -1936,7 +2477,6 @@ static bool consume_llc(struct llc_ctx *llcx)
 					 taskc->enq_flags);
 		bpf_task_release(p);
 
-		/* Try to dispatch from LLC DSQ (handles affinity check atomically) */
 		return scx_bpf_dsq_move_to_local(cpuc->llc_dsq);
 	}
 try_dsq:
@@ -2231,10 +2771,6 @@ check_llc_dsq:
 		trace("ATQ dispatching %llu with min vtime %llu", min_atq, min_vtime);
 		pid = scx_atq_pop(min_atq);
 		if (likely((p = bpf_task_from_pid((s32)pid)))) {
-			/*
-			 * the ATQ. Otherwise there may be priority inversions.
-			 * This probably needs to be done for the DSQs as well.
-			 */
 			if (unlikely(!(taskc = lookup_task_ctx(p)))) {
 				bpf_task_release(p);
 				scx_bpf_error("failed to get task ctx");
@@ -2242,7 +2778,6 @@ check_llc_dsq:
 			}
 
 
-			/* Insert to LLC DSQ for atomic affinity handling */
 			scx_bpf_dsq_insert_vtime(p,
 						 cpuc->llc_dsq,
 						 taskc->slice_ns,
@@ -2250,7 +2785,6 @@ check_llc_dsq:
 						 taskc->enq_flags);
 			bpf_task_release(p);
 
-			/* Try to dispatch - move_to_local handles affinity atomically */
 			scx_bpf_dsq_move_to_local(cpuc->llc_dsq);
 			return;
 		}
@@ -2297,9 +2831,6 @@ check_llc_dsq:
 				return;
 			}
 
-			/* Check if task can still run on current CPU */
-
-			/* Insert to LLC DSQ for atomic affinity handling */
 			scx_bpf_dsq_insert_vtime(p,
 						 cpuc->llc_dsq,
 						 taskc->slice_ns,
@@ -2307,7 +2838,6 @@ check_llc_dsq:
 						 taskc->enq_flags);
 			bpf_task_release(p);
 
-			/* Try to dispatch - move_to_local handles affinity atomically */
 			scx_bpf_dsq_move_to_local(cpuc->llc_dsq);
 		}
 	} else if (unlikely(p2dq_config.atq_enabled)) {
@@ -2319,9 +2849,6 @@ check_llc_dsq:
 				return;
 			}
 
-			/* Check if task can still run on current CPU */
-
-			/* Insert to LLC DSQ for atomic affinity handling */
 			scx_bpf_dsq_insert_vtime(p,
 						 cpuc->llc_dsq,
 						 taskc->slice_ns,
@@ -2329,7 +2856,6 @@ check_llc_dsq:
 						 taskc->enq_flags);
 			bpf_task_release(p);
 
-			/* Try to dispatch - move_to_local handles affinity atomically */
 			scx_bpf_dsq_move_to_local(cpuc->llc_dsq);
 			return;
 		}
@@ -2339,14 +2865,12 @@ check_llc_dsq:
 			return;
 	}
 
-	// Lookup LLC ctx (should never fail at this point)
 	if (unlikely(p2dq_config.llc_shards <= 1 &&
 	    !(llcx = lookup_llc_ctx(cpuc->llc_id)))) {
 		scx_bpf_error("invalid llc id %u", cpuc->llc_id);
 		return;
 	}
 
-	// Try to keep prev task running (optimization for low-latency tasks)
 	if (unlikely(prev && keep_running(cpuc, llcx, prev)))
 		return;
 
@@ -2544,6 +3068,14 @@ static s32 p2dq_init_task_impl(struct task_struct *p, struct scx_init_task_args
 	else
 		taskc->dsq_id = cpuc->llc_dsq;
 
+	taskc->pid = p->pid;
+
+	if (p->flags & PF_FORKNOEXEC)
+		task_ctx_set_flag(taskc, TASK_CTX_F_FORKNOEXEC);
+	else
+		task_ctx_clear_flag(taskc, TASK_CTX_F_FORKNOEXEC);
+	taskc->target_llc_hint = MAX_LLCS;
+
 	return 0;
 }
 
@@ -3053,6 +3585,50 @@ void BPF_STRUCT_OPS(p2dq_exit, struct scx_exit_info *ei)
 	UEI_RECORD(uei, ei);
 }
 
+/*
+ * Thermal Pressure Tracking (requires CONFIG_SCHED_HW_PRESSURE=y)
+ *
+ * Thermal tracking provides optimization by avoiding throttled CPUs.
+ * This program has autoload disabled by default and is conditionally
+ * enabled from userspace if the kernel supports hw_pressure_update tracepoint.
+ *
+ * The '?' suffix makes this program optional - veristat and the verifier
+ * will skip it if the tracepoint doesn't exist in the kernel.
+ *
+ * Tracepoint: hw_pressure_update
+ * Fires when kernel detects thermal throttling on a CPU
+ *
+ * Arguments:
+ *   cpu: CPU ID experiencing pressure
+ *   hw_pressure: Pressure value (0 = no throttling, 1024 = max capacity lost)
+ *
+ * Note: This tracepoint only exists on ARM/ARM64 architectures
+ */
+#if defined(__TARGET_ARCH_arm64) || defined(__aarch64__)
+__weak __hidden SEC("tp_btf/hw_pressure_update?")
+int BPF_PROG(on_thermal_pressure, u32 cpu, u64 hw_pressure)
+{
+	struct cpu_ctx *cpuc;
+
+	if (cpu >= MAX_CPUS)
+		return 0;
+
+	cpuc = lookup_cpu_ctx(cpu);
+	if (!cpuc)
+		return 0;
+
+	cpuc->perf = (u32)hw_pressure;
+
+	if (hw_pressure > 512) {
+		scx_bpf_kick_cpu(cpu, SCX_KICK_IDLE);
+		stat_inc(P2DQ_STAT_THERMAL_KICK);
+	}
+
+	return 0;
+}
+#endif
+
+
 #if P2DQ_CREATE_STRUCT_OPS
 s32 BPF_STRUCT_OPS_SLEEPABLE(p2dq_init)
 {
@@ -3074,9 +3650,13 @@ void BPF_STRUCT_OPS(p2dq_enqueue, struct task_struct *p __arg_trusted, u64 enq_f
 
 void BPF_STRUCT_OPS(p2dq_dequeue, struct task_struct *p __arg_trusted, u64 deq_flags)
 {
-	task_ctx *taskc = lookup_task_ctx(p);
+	task_ctx *taskc;
 	int ret;
 
+	if (!p2dq_config.atq_enabled)
+		return;
+
+	taskc = lookup_task_ctx(p);
 	ret = scx_atq_cancel(&taskc->common);
 	if (ret)
 		scx_bpf_error("scx_atq_cancel returned %d", ret);
