@@ -5,11 +5,10 @@
  */
 
 #include <scx/common.bpf.h>
-#include <scx/ravg_impl.bpf.h>
 #include <lib/sdt_task.h>
 
-#include <scx/bpf_arena_common.bpf.h>
-#include <scx/bpf_arena_spin_lock.h>
+#include <bpf_arena_common.bpf.h>
+#include <bpf_arena_spin_lock.h>
 
 #include <lib/cpumask.h>
 
