@@ -18,6 +18,10 @@ pub struct TraceCapabilities {
     pub available_events: HashSet<String>,
     /// Event type -> count
     pub event_counts: HashMap<String, usize>,
+    /// TrackEvent categories found in trace (from wprof or other TrackEvent sources)
+    pub track_event_categories: HashSet<String>,
+    /// TrackEvent category -> count
+    pub track_event_counts: HashMap<String, usize>,
     /// Clock sources present in trace
     pub clock_sources: HashSet<ClockType>,
     /// Has ProcessTree packets
@@ -42,6 +46,27 @@ pub struct TraceCapabilities {
     pub trace_source: TraceSource,
 }
 
+/// Maps wprof TrackEvent categories to equivalent ftrace event types
+/// This allows analyzers to work with both ftrace and TrackEvent data
+pub fn map_wprof_category_to_ftrace(category: &str) -> Option<&'static str> {
+    match category {
+        "ONCPU" => Some("sched_switch"),
+        "WAKEE" | "WAKEE_NEW" => Some("sched_wakeup"),
+        "WAKER" | "WAKER_NEW" => Some("sched_waking"),
+        "PREEMPTEE" | "PREEMPTOR" => Some("sched_switch"), // preemption is part of sched_switch
+        "HARDIRQ" => Some("irq_handler_entry"),
+        "SOFTIRQ" => Some("softirq_entry"),
+        "FORKING" | "FORKED" => Some("sched_process_fork"),
+        "EXIT" => Some("sched_process_exit"),
+        "EXEC" => Some("sched_process_exec"),
+        "START" => Some("sched_process_exec"),
+        "RENAME" => None, // No direct ftrace equivalent
+        "FREE" => None,   // No direct ftrace equivalent
+        "WQ" => None,     // Workqueue events
+        _ => None,
+    }
+}
+
 impl TraceCapabilities {
     /// Build capabilities from a parsed trace
     pub fn from_trace(trace: &PerfettoTrace) -> Self {
@@ -87,13 +112,36 @@ impl TraceCapabilities {
             }
         }
 
+        // Get TrackEvent categories (used by wprof and other TrackEvent-based traces)
+        let track_event_counts = trace.get_track_event_counts_by_category();
+        let track_event_categories: HashSet<String> = track_event_counts.keys().cloned().collect();
+
+        // Map TrackEvent categories to equivalent ftrace events for analyzer compatibility
+        // This allows analyzers designed for ftrace to also work with TrackEvent data
+        for category in &track_event_categories {
+            if let Some(ftrace_equiv) = map_wprof_category_to_ftrace(category) {
+                available_events.insert(ftrace_equiv.to_string());
+                // Add the TrackEvent count to the ftrace equivalent
+                if let Some(count) = track_event_counts.get(category) {
+                    *event_counts.entry(ftrace_equiv.to_string()).or_insert(0) += count;
+                }
+            }
+        }
+
         // Check for process tree
         let has_process_tree = !trace.get_processes().is_empty();
 
         // Detect trace source based on available data
+        // wprof traces have characteristic categories like ONCPU, WAKEE, WAKER
+        let is_wprof = track_event_categories.contains("ONCPU")
+            && (track_event_categories.contains("WAKEE")
+                || track_event_categories.contains("WAKER"));
+
         let trace_source = if has_sched_ext {
             TraceSource::Scxtop
-        } else if available_events.contains("sched_switch") {
+        } else if is_wprof {
+            TraceSource::Wprof
+        } else if available_events.contains("sched_switch") && track_event_categories.is_empty() {
             TraceSource::GenericFtrace
         } else {
             TraceSource::Unknown
@@ -102,6 +150,8 @@ impl TraceCapabilities {
         Self {
             available_events,
             event_counts,
+            track_event_categories,
+            track_event_counts,
             clock_sources,
             has_process_tree,
             has_system_info,
@@ -326,6 +376,8 @@ impl ClockType {
 pub enum TraceSource {
     /// Generated by scxtop
     Scxtop,
+    /// Generated by wprof
+    Wprof,
     /// Android systrace
     AndroidSystemTrace,
     /// Chrome tracing
@@ -340,6 +392,7 @@ impl TraceSource {
     pub fn as_str(&self) -> &'static str {
         match self {
             TraceSource::Scxtop => "scxtop",
+            TraceSource::Wprof => "wprof",
             TraceSource::AndroidSystemTrace => "Android Systrace",
             TraceSource::ChromeTracing => "Chrome Tracing",
             TraceSource::GenericFtrace => "Generic Ftrace",
@@ -437,6 +490,8 @@ mod tests {
         let mut caps = TraceCapabilities {
             available_events: HashSet::new(),
             event_counts: HashMap::new(),
+            track_event_categories: HashSet::new(),
+            track_event_counts: HashMap::new(),
             clock_sources: HashSet::new(),
             has_process_tree: true,
             has_system_info: false,
@@ -464,6 +519,8 @@ mod tests {
         let mut caps = TraceCapabilities {
             available_events: HashSet::new(),
             event_counts: HashMap::new(),
+            track_event_categories: HashSet::new(),
+            track_event_counts: HashMap::new(),
             clock_sources: HashSet::new(),
             has_process_tree: true,
             has_system_info: false,
